/**
 * delight.BIH
 * 
 * License: GPL
 * Author: Pyo Taesu
 */
module delight.BIH;
private import delight.Math;
private import delight.Util;
private import delight.Ray;
private import tango.core.Array;
private import tango.io.Stdout;

/**
 * Bounding Interval Hierarchy
 * 
 * Refereces:
 *   http://ainc.de/Research/BIH.pdf
 *   http://en.wikipedia.org/wiki/Bounding_interval_hierarchy
 */
class BIH
{
	struct Node
	{
		uint children_axis;
		float[2] sep;
		
		uint childIndex() { return children_axis >> 2; }
		uint axis()
		{
			// 0 : x-axis
			// 1 : y-axis
			// 2 : z-axis
			// 3 : leaf node
			return children_axis & 0x03;
		}
		
		void init(uint nodeIndex)
		{
			// leaf node
			children_axis = (nodeIndex << 2) | 3;
		}
		
		void init(uint childIndex, uint axis, float sep1, float sep2)
		{
			// internal node
			children_axis = (childIndex << 2) | axis;
			sep[0] = sep1;
			sep[1] = sep2;
		}
	}
	
	Node[] nodes;
	AABB aabb;
	
	void build(AABB[] boundingBoxes)
	{
		Stdout("buildBIH: ")(boundingBoxes.length)(" nodes...").newline.flush;
		
		if (boundingBoxes.length == 0)
		{
			nodes = null;
			aabb = AABB(Vector3(0), Vector3(0));
			return;
		}
		
		nodes = new Node[boundingBoxes.length * 2 - 1];
		auto indexes = new uint[boundingBoxes.length];
		auto nodeCount = 0;
		
		foreach(uint i, ref uint v ; indexes)
		{
			v = i;
		}
		
		void buildSub(uint nodeIndex, uint[] subIndexes, ref AABB aabb)
		{
			if (subIndexes.length == 1)
			{
				// leaf node
				auto index = subIndexes[0];
				aabb = boundingBoxes[index];
				nodes[nodeIndex].init(index);
			}
			else
			{
				// internal node
				assert(subIndexes.length > 1);
				
				// update bounding box
				aabb = boundingBoxes[subIndexes[0]];
				
				foreach(index ; subIndexes[1..$])
				{
					aabb.add(boundingBoxes[index]);
				}
				
				// find longest axis
				auto axis = aabb.size.maxAxis;
				auto sepVal = aabb.center[axis];

				float centerVal(uint i)
				{
					return boundingBoxes[i].center[axis];
				}
				
				// partition by sepVal
				uint count;
				count = partition(subIndexes,
					(uint index) { return centerVal(index) < sepVal; }
					);
				
				if (count < 1)
				{
					// partition by sorted order
					sort(subIndexes,
						(uint i1, uint i2) { return centerVal(i1) < centerVal(i2); }
						);
					count = subIndexes.length / 2;
				}
				
				auto child1 = nodeCount;
				auto child2 = nodeCount+1;
				nodeCount += 2;
				
				AABB aabb1, aabb2;
				
				buildSub(child1, subIndexes[0..count], aabb1);
				buildSub(child2, subIndexes[count..$], aabb2);

				nodes[nodeIndex].init(child1, axis, aabb1.v2[axis], aabb2.v1[axis]);
			}
		}
		
		++nodeCount;
		buildSub(0, indexes, aabb);
		
		assert(nodes.length == nodeCount);
	}
	
	struct TraceRayForEach
	{
		BIH bih;
		Ray * ray;
		
		int opApply(int delegate(ref uint) dg)
		{
			return bih.traceRay(ray, dg);
		}
	}
	
	TraceRayForEach traceRay(Ray * ray)
	{
		TraceRayForEach it = { this, ray };
		return it;
	}
	
	int traceRay(Ray * ray, int delegate(ref uint) dg)
	{
		int r = 0;
		if (nodes)
		{
			r = traceRay(aabb, 0, ray, dg);
		}
		return r;
	}
	
	int traceRay(AABB aabb, uint nodeIndex, Ray * ray, int delegate(ref uint) dg)
	{
		int r = 0;
		if (ray.intersects(aabb))
		{
			Node n = nodes[nodeIndex];
			
			uint axis = n.axis; 
			if (3 == axis)
			{
				// leaf
				uint itemIndex = n.childIndex;
				r = dg(itemIndex);
			}
			else
			{
				// internal node
				int i1 = 0;
				int i2 = 1;
				
				if (ray.direction[axis] < 0)
				{
					// reverse direction
					swap(i1, i2);
				}

				AABB[2] childAabb = void;
				childAabb[0] = aabb;
				childAabb[0].v2[axis] = n.sep[0];
				childAabb[1] = aabb;
				childAabb[1].v1[axis] = n.sep[1];
				
				r = traceRay(childAabb[i1], n.childIndex+i1, ray, dg);
				if (0 == r)
				{
					r = traceRay(childAabb[i2], n.childIndex+i2, ray, dg);
				}
			}
		}
		
		return r;
	}
}
