/**
 * delight.math module
 * 
 * License: GPL
 * Author: Pyo Taesu
 */
module delight.Math;

private import Math = tango.math.Math;
private import tango.text.convert.Format;

alias Math.abs abs;
alias Math.pow pow;
alias Math.sqrt sqrt;

float frac(float v)
{
	return v - Math.floor(v);
}

float degreeToRadian(float v)
{
	return v / 180.f * Math.PI;
}

T neg1(T)(T a) { return -a; }
T add2(T)(T a, T b) { return a + b; }
T sub2(T)(T a, T b) { return a - b; }
T mul2(T,U)(T a, U b) { return a * b; }

T clamp(T)(T v, T minVal, T maxVal)
{
	return Math.min(Math.max(v, minVal), maxVal);
}

float saturate(float v)
{
	return clamp!(float)(v, 0, 1);
}

T lerp(T)(T a, T b, float t)
{
	return a + (b - a) * t;
}

template applyX(T)
{
	T apply1(alias f)()
	{
		T r = void;
		foreach(i, u ; this.tupleof)
		{
			r.tupleof[i] = f(u);
		}
		return r;
	}

	T apply2(alias f)(T v)
	{
		T r = void;
		foreach(i, u ; this.tupleof)
		{
			r.tupleof[i] = f(u, v.tupleof[i]);
		}
		return r;
	}

	T apply2k(alias f)(float v)
	{
		T r = void;
		foreach(i, u ; this.tupleof)
		{
			r.tupleof[i] = f(u, v);
		}
		return r;
	}

	void apply2Assign(alias f)(T v)
	{
		foreach(i, u ; this.tupleof)
		{
			this.tupleof[i] = f(u, v.tupleof[i]);
		}
	}

	void apply2kAssign(alias f)(float v)
	{
		foreach(i, u ; this.tupleof)
		{
			this.tupleof[i] = f(u, v);
		}
	}
}

template opXXX1(alias T)
{
	T opNeg()
	{
		return apply1!(neg1)();
	}
}

template opXXX2(alias T)
{
	T opAdd(T v)
	{		
		return apply2!(add2)(v);
	}

	void opAddAssign(T v)
	{		
		apply2Assign!(add2)(v);
	}

	T opSub(T v)
	{
		return apply2!(sub2)(v);
	}

	void opSubAssign(T v)
	{
		apply2Assign!(sub2)(v);
	}

	T opMul(T v)
	{
		return apply2!(mul2)(v);
	}

	void opMulAssign(T v)
	{
		apply2Assign!(mul2)(v);
	}

	T opMul(float k)
	{
		return apply2k!(mul2)(k);
	}

	void opMulAssign(float k)
	{
		apply2kAssign!(mul2)(k);
	}

	T opDiv(float k)
	{
		return opMul(1.0f / k);
	}

	void opDivAssign(float k)
	{
		opMulAssign(1.0f / k);
	}
}

template opXXXR(alias T)
{
	float sum()
	{
		return reduce!(add2)();
	}
	
	float min()
	{
		return reduce!(Math.min)();
	}
	
	float max()
	{
		return reduce!(Math.max)();
	}
}

/**
 * 3D vector implementation 
 */
struct Vector3
{
	float x = 0, y = 0, z = 0;
	
	float r() { return x; }
	float g() { return y; }
	float b() { return z; }
	
	float opIndex(uint i) { return (&x)[i]; }
	void opIndexAssign(float v, uint i) { (&x)[i] = v; }

	static Vector3 opCall(float x, float y, float z)
	{
		Vector3 r = { x, y, z, };
		return r;
	}

	static Vector3 opCall(float v)
	{
		return Vector3(v, v, v);
	}
	
	static Vector3 opCall(float[] v)
	{
		Vector3 r;
		foreach( i, f ; v[0..3])
		{
			r[i] = f;
		}
		return r;
	}
	
	float reduce(alias f)()
	{
		return f(f(x, y), z);
	}

	mixin applyX!(Vector3);
	mixin opXXX1!(Vector3);
	mixin opXXX2!(Vector3);
	mixin opXXXR!(Vector3);
	
	Vector3 yzx() { return Vector3(y, z, x); }
	Vector3 zxy() { return Vector3(z, x, y); }
	
	uint maxAxis()
	{
		if (z > y)
		{
 			if (z > x) return 2;
		}
		else
		{
			if (y > x) return 1;
		}
		
		return 0;
	}

	char[] toString()
	{
		return Format.convert("({0}, {1}, {2})", x, y, z);
	}
}

/**
 * 3x4 matrix implementation 
 */
struct Matrix34
{
	Vector3 col0, col1, col2, col3;
	
	static Matrix34 opCall(Vector3 c0, Vector3 c1, Vector3 c2, Vector3 c3 = Vector3(0))
	{
		Matrix34 r = void;
		r.col0 = c0;
		r.col1 = c1;
		r.col2 = c2;
		r.col3 = c3;
		return r;
	}

	static Matrix34 opCall(float v)
	{
		return scale(v,v,v);
	}
	
	static Matrix34 opCall(float[] v)
	{
		Matrix34 r = identity;
		r.col0.x = v[0];
		r.col1.x = v[1];
		r.col2.x = v[2];
		r.col3.x = v[3];
		r.col0.y = v[4];
		r.col1.y = v[5];
		r.col2.y = v[6];
		r.col3.y = v[7];
		r.col0.z = v[8];
		r.col1.z = v[9];
		r.col2.z = v[10];
		r.col3.z = v[11];
		return r;
	}
	
	static Matrix34 identity()
	{
		return scale(1,1,1);
	}
	
	static Matrix34 translation(Vector3 v)
	{
		return Matrix34(Vector3(1,0,0), Vector3(0,1,0), Vector3(0,0,1), v);
	}

	static Matrix34 translation(float x, float y, float z)
	{
		return translation(Vector3(x, y, z));
	}

	static Matrix34 scale(float x, float y, float z)
	{
		return Matrix34(Vector3(x,0,0), Vector3(0,y,0), Vector3(0,0,z));		
	}
	
	Vector3 row0() { return Vector3(col0.x, col1.x, col2.x); }
	Vector3 row1() { return Vector3(col0.y, col1.y, col2.y); }
	Vector3 row2() { return Vector3(col0.z, col1.z, col2.z); }

	mixin applyX!(Matrix34);
	mixin opXXX1!(Matrix34);
	mixin opXXX2!(Matrix34);
	
	float determinant()
	{
		return dot(col0, cross(col1, col2));
	}
	
	Matrix34 inverse()
	{
		Vector3 v0 = cross(row1, row2);
		Vector3 v1 = cross(row2, row0);
		Vector3 v2 = cross(row0, row1);
		auto det = dot(row0, v0);
		
		auto m = Matrix34(v0, v1, v2) / det;
		m.col3 = -rotate(m, col3);
		
		return m;
	}
	
	char[] toString()
	{
		return Format.convert("({0}, {1}, {2}, {3})", col0, col1, col2, col3);
	}
}

Vector3 min(Vector3 a, Vector3 b)
{
	return a.apply2!(Math.min)(b);
}

Vector3 max(Vector3 a, Vector3 b)
{
	return a.apply2!(Math.max)(b);
}

T clamp(T:Vector3)(T v, T minVal, T maxVal)
{
	return min(max(v, minVal), maxVal);
}

Vector3 saturate(Vector3 v)
{
	return clamp(v, Vector3(0), Vector3(1));
}

float dot(Vector3 a, Vector3 b)
{
	return (a * b).sum;
}

float lengthSq(Vector3 a)
{
	return dot(a, a);
}

float length(Vector3 a)
{
	return Math.sqrt(dot(a, a));
}

Vector3 normalize(Vector3 a)
{
	return a / length(a);
}

/**
 * projection of a on b
 */
Vector3 projection(Vector3 a, Vector3 b)
{
	return b * (dot(a, b) / dot(b, b));
}

Vector3 orthogonalize(Vector3 a, Vector3 b)
{
	return a - projection(a, b);
}

/**
 * reflection of a on b
 */
Vector3 reflect(Vector3 a, Vector3 b)
{
	return a - projection(a, b) * 2;
}

/**
 * corss product
 * 
 * [ i  j  k  ]
 * [ ax ay az ]
 * [ bx by bz ]
 * 
 * (ay * bz, az * bx, ax * by) - (az * by, ax * bz, ay * bx)
 */
Vector3 cross(Vector3 a, Vector3 b)
{
	return a.yzx * b.zxy - a.zxy * b.yzx;
}

Vector3 rotate(Matrix34 m, Vector3 v)
{
	return m.col0 * v.x + m.col1 * v.y + m.col2 * v.z;
}

Vector3 mul(Matrix34 m, Vector3 v)
{
	return rotate(m, v) + m.col3;
}

Matrix34 mul(Matrix34 a, Matrix34 b)
{
	return Matrix34(
		rotate(a, b.col0),
		rotate(a, b.col1),
		rotate(a, b.col2),
		mul(a, b.col3));
}

Vector3 direction(float th)
{
	return Vector3(Math.cos(th), Math.sin(th), 0.0f);
}

/**
 * Axis-Aligned Bounding Box
 */
struct AABB
{
	Vector3 v1, v2;
	
	static AABB opCall(Vector3 v1, Vector3 v2)
	{
		AABB aabb = void;
		aabb.v1 = v1;
		aabb.v2 = v2;
		return aabb;
	}

	static AABB opCall(Vector3 v)
	{
		return opCall(v, v);
	}

	void assign(Vector3 v)
	{
		v1 = v;
		v2 = v;
	}
	
	void add(Vector3 v)
	{
		v1 = min(v1, v);
		v2 = max(v2, v);
	}
	
	void add(AABB aabb)
	{
		add(aabb.v1);
		add(aabb.v2);
	}
	
	Vector3 size()
	{
		return v2 - v1;
	}
	
	Vector3 center()
	{
		return (v1 + v2) /2;
	}
	
	char[] toString()
	{
		return Format.convert("({0}, {1})", v1, v2);
	}
}
