/**
 * delight.Mesh
 * 
 * License: GPL
 * Author: Pyo Taesu
 */
module delight.Mesh;
private import delight.Math;
private import delight.BIH;
private import delight.Util;
private import delight.Ray;
private import tango.io.stream.FileStream;
private import tango.io.stream.LineStream;
private import tango.io.Stdout;
private import tango.text.Util;
private import tango.util.collection.ArraySeq;
private import Math = tango.math.Math;
private import Float = tango.text.convert.Float;
private import Integer = tango.text.convert.Integer;

/**
 * Triangle mesh
 */
class Mesh
{
	char[] name = "";
	BIH bvh; // bounding volume hierarchy
	Vector3[] vertPositions;
	Vector3[] vertNormals;
	Face[] faces;
	
	this()
	{
	}
	
	AABB getAABB(Matrix34 worldTransform)
	{
		auto aabb = AABB(mul(worldTransform, vertPositions[0]));
		
		foreach( i ; range(1u, vertPositions.length))
		{
			aabb.add(mul(worldTransform, vertPositions[i]));
		}
		
		return aabb;
	}
	
	void flipTriangles()
	{
		foreach(ref Face f ; faces)
		{
			f.flip();
		}
	}
	
	void buildPlane()
	{
		vertPositions = new Vector3[4];
		vertNormals = new Vector3[1];
		faces = new Face[2];
		
		foreach(i ; range(4))
		{
			vertPositions[i] = direction((i + 0.5f) * Math.PI / 2);
		}
		
		vertNormals[0] = Vector3(0,0,1);
		
		faces[0].vertPositions = [0, 1, 2];
		faces[0].vertNormals = [0, 0, 0];
		faces[1].vertPositions = [2, 3, 0];
		faces[1].vertNormals = [0, 0, 0];
	}

	void loadFromObjFile(char[] fileName)
	{
		Stdout("loading: ")(fileName)("...").newline.flush;
		
		// Reference : http://en.wikipedia.org/wiki/Obj
		
		scope file = new FileInput(fileName);
		
		scope vlist = new ArraySeq!(Vector3);
		scope vnlist = new ArraySeq!(Vector3);
		scope flist = new ArraySeq!(Face);
		
		vlist.capacity(1000);
		vnlist.capacity(1000);
		flist.capacity(1000);
		
		foreach(s ; new LineInput(file))
		{
			if (0 == s.length || s[0] == '#') continue;
			
			if (startsWith(s, "v "))
			{
				// s <= "v 1.189148 1.189147 -1.189148"
				
				Vector3 v;
				foreach(i, e ; split(s, " ")[1..4])
				{
					v[i] = Float.parse(e);
				}

				vlist.append(v);
			}
			else if (startsWith(s, "vn "))
			{
				Vector3 v;
				foreach(i, e ; split(s, " ")[1..4])
				{
					v[i] = Float.parse(e);
				}

				vnlist.append(v);
			}
			else if (startsWith(s, "f "))
			{
				// s <= "f 5//1 1//1 4//1" (1-based index)
				
				Face f;
				
				foreach(i, e ; split(s, " ")[1..4])
				{
					auto e2 = split(e, "/")[0..3];
					foreach(j, v ; e2)
					{
						f.vertsArray[j][i] = Integer.parse(e2[j]) - 1;
					}
				}
				
				flist.append(f);
			}
		}
		
		vertPositions = vlist.toArray;
		vertNormals = vnlist.toArray;
		faces = flist.toArray;
		
		buildBVH();
	}
	
	void buildBVH()
	{
		scope boundingBoxes = new AABB[faces.length];
		
		foreach(i, f ; faces)
		{
			AABB * aabb = &boundingBoxes[i];
			
			aabb.assign(vertPositions[f.vertPositions[0]]);
			aabb.add(vertPositions[f.vertPositions[1]]);
			aabb.add(vertPositions[f.vertPositions[2]]);
		}
		
		bvh = new BIH;
		bvh.build(boundingBoxes);
	}
	
	void traceRay(RayContact * rc)
	{
		if (bvh)
		{
			foreach( i ; bvh.traceRay(&rc.ray) )
			{
				intersectsRayFace(rc, i);
			}
		}
		else
		{
			foreach( i ; range(faces.length) )
			{
				intersectsRayFace(rc, i);
			}
		}
	}
	
	void intersectsRayFace(RayContact * rc, uint faceIndex)
	{
		// Ray-Triangle Intersection
		
		Ray ray = rc.ray;
		Vector3[3] v = void;
		auto face = &faces[faceIndex];
		
		foreach( i, vi ; face.vertsArray[Face.POSITION] )
		{
			v[i] = vertPositions[vi] - ray.origin;
		}
		
		// ray.direction * t = v0 * a + v1 * b + v2 * (1 - a - b)
		// (t > 0, 0 <= a <= 1, 0 <= b <= 1)
		
		auto c1 = ray.direction;
		auto c2 = v[2] - v[0];
		auto c3 = v[2] - v[1];
		
		// c1 * t + c2 * a + c3 * b = v2
		
		// [          ] [ t ]
		// [ c1 c2 c3 ] [ a ] = v2
		// [          ] [ b ]
		
		auto r1 = cross(c2, c3);
		auto r2 = cross(c3, c1);
		auto r3 = cross(c1, c2);
		auto det = dot(c1, r1);
		
		// [ t ]             [   r1   ]
		// [ a ] = (1/det) * [   r2   ] v2
		// [ b ]             [   r3   ]
	
		// det = dot(ray, normal) * k
		// k > 0
		if (det < 0)
		{
			auto t = dot(r1, v[2]) / det;
			
			if (ray.insideRange(t))
			{
				auto a = dot(r2, v[2]) / det;
				auto b = dot(r3, v[2]) / det;
				if (a >= 0 && b >= 0 && a + b <= 1)
				{
					rc.ray.maxDistance = t;
					rc.distance = t;
					rc.normal = face.normal(vertNormals, a, b);
				}
			}
		}
	}
}

struct Face
{
	enum
	{
		POSITION = 0,
		NORMAL = 2,
	}
	
	int[3][3] vertsArray;
	
	int * vertPositions() {  return &vertsArray[POSITION][0]; }
	void vertPositions(int[] a) { assignVertArray(POSITION, a); }

	int * vertNormals() {  return &vertsArray[NORMAL][0]; }
	void vertNormals(int[] a) { assignVertArray(NORMAL, a); }

	void assignVertArray(uint arrIndex, int[] a)
	{
		foreach( i ; range(vertsArray[arrIndex].length) )
		{
			vertsArray[arrIndex][i] = a[i];
		}
	}
	
	Vector3 normal(Vector3[] normals, float a, float b)
	{
		return normalize(
			normals[vertNormals[0]] * a +
			normals[vertNormals[1]] * b + 
			normals[vertNormals[2]] * (1 - a - b)); 
	}
	
	void flip()
	{
		foreach( i ; range(vertsArray.length) )
		{
			swap(vertsArray[i][1], vertsArray[i][1]);
		}
	}
}
