/**
 * delight.Renderer
 * 
 * License: GPL
 * Author: Pyo Taesu
 */
module delight.Renderer;
private import delight.Scene;
private import delight.Image;
private import delight.Util;
private import delight.Ray;
private import delight.Math;
private import delight.Material;
private import tango.io.Stdout;

struct RenderBucket
{
	int left, top;
	int right, bottom;
	
	Range!(int) xrange()
	{
		return range(left, right);
	}

	Range!(int) yrange()
	{
		return range(top, bottom);
	}
}

class Renderer
{
	Scene scene;
	Image image;
	
	this(Scene s, Image img)
	{
		scene = s;
		image = img;
		image.initImage(scene.renderingOption.screenWidth, scene.renderingOption.screenHeight);
	}
	
	void render()
	{
		Stdout("Rendering (")(image.width)("x")(image.height)(")")("...").newline.flush;
		
		RenderBucket rb;
		rb.left = 0;
		rb.top = 0;
		rb.right = image.width;
		rb.bottom = image.height;
		
		render(&rb);
	}
	
	void render(RenderBucket * bk)
	{
		auto camera = scene.camera;
		Matrix34 transform;
		
		with(Matrix34)
		{
			float w = image.width;
			float h = image.height;
			float t = Math.tan(camera.fov/2);
			transform = mul(scale(t/h,-t/h,1), translation(-w/2,-h/2,-1));
		}
		
		void processPixel(int x, int y)
		{
			auto localRayDir = mul(transform, Vector3(x, y, 0));
			image.setPixel(x, y, traceRay(camera.pickRay(localRayDir), scene.renderingOption.traceDepth));
		}

		foreach(y ; bk.yrange)
		{
			foreach(x ; bk.xrange)
			{
				processPixel(x, y);
			}
		}
	}
	
	Vector3 traceRay(Ray ray, int traceDepth)
	{
		RayContact contact;
		contact.ray = ray;
		scene.traceRay(&contact);
		
		if (contact.object)
		{
			return shadeSurface(contact.object, contact.contactPosition, contact.normal, ray.direction, traceDepth);
		}
		
		return Vector3(0);
	}
	
	Vector3 shadeSurface(SceneObject obj, Vector3 pos, Vector3 normal, Vector3 V, int traceDepth)
	{
		auto color = Vector3(0);
		
		ShadingParams params;
		params.N = normal;
		params.V = V;
		params.R = reflect(V, normal);
		
		foreach( light ; scene.lights )
		{
			params.L = light.position - pos;
			
			if (dot(params.L, normal) <= 0) continue;
			
			auto dist = length(params.L);
			
			if (dist < 0.000001f) continue;
			
			params.L /= dist;
			
			if (!checkOcclusion(Ray(pos, params.L)))
			{
				params.CL = light.lightColor;
				color += obj.material.shade(&params);
			}
		}
		
		if (traceDepth > 0)
		{
			if (obj.material.refractive > 0.0001f)
			{
				auto f = obj.material.refractive;
				auto v2 = obj.material.refract(V, normal, 1.0f);
				auto c2 = traceRay(Ray(pos, v2, 0.0001f), traceDepth - 1);
				color = lerp(color, c2, f);
			}
		
			if (obj.material.reflective > 0.0001f)
			{
				auto cosi = -dot(V, normal);
				auto etai = 1.0f;
				float f = obj.material.fresnelReflectance(etai, cosi) * obj.material.reflective;
				
				if (f > 0.0001f)
				{
					auto c2 = traceRay(Ray(pos, params.R), traceDepth - 1);
					color = lerp(color, c2, f);
				}
			}
		}
		
		return color;
	}
	
	bool checkOcclusion(Ray ray)
	{
		RayContact ct;
		ct.ray = ray;
		scene.traceRay(&ct);
		
		return ct.object !is null;
	}
}
