/**
 * delight.Scene
 * 
 * License: GPL
 * Author: Pyo Taesu
 */
module delight.Scene;
private import delight.Math;
private import delight.Mesh;
private import delight.Material;
private import delight.Camera;
private import delight.Light;
private import delight.BIH;
private import delight.Ray;
private import tango.io.Stdout;


struct RenderingOption
{
	uint screenWidth = 800, screenHeight = 600;
	uint traceDepth = 3;
}

class Scene
{
	SceneObject[] objects = [];
	Light[] lights = [];
	MaterialObject[] materials = [];
	Mesh[] meshes = [];
	Camera camera;
	BIH bvh;
	RenderingOption renderingOption;
	
	this()
	{
	}

	void addLight(Light obj)
	{
		lights ~= [obj];
	}

	void addMaterial(char[] name, Material material)
	{
		materials ~= [new MaterialObject(name, material)];
	}
	
	Material getMaterial(char[] name)
	{
		foreach(o ; materials)
		{
			if (o.name == name) return o.material;
		}
		return Material.init;
	}
	
	void addMesh(Mesh obj)
	{
		meshes ~= [obj];
	}
	
	Mesh getMesh(char[] name)
	{
		foreach( m ; meshes )
		{
			if (m.name == name) return m;
		}
		return null;
	}
	
	void addObject(SceneObject obj)
	{
		objects ~= [obj];
	}

	void buildBVH()
	{
		if (bvh is null) bvh = new BIH;
		
		auto boundingBoxes = new AABB[objects.length];
		
		foreach( i, obj ; objects )
		{
			boundingBoxes[i] = obj.getAABB();
		}
		
		bvh.build(boundingBoxes);
	}
	
	void traceRay(RayContact * rc)
	{
		if (bvh)
		{
			foreach( i ; bvh.traceRay(&rc.ray) )
			{
				objects[i].traceRay(rc);
			}
		}
		else
		{
			foreach( obj ; objects)
			{
				obj.traceRay(rc);
			}
		}
	}
}

class SceneObjectBase
{
	char[] name = "";
	
	Matrix34 _worldTransform;
	Matrix34 _invWorldTransform;
	
	Matrix34 worldTransform() { return _worldTransform; }
	void worldTransform(Matrix34 v)
	{
		_worldTransform = v;
		_invWorldTransform = v.inverse;
	}
	
	Matrix34 invWorldTransform() { return _invWorldTransform; }
	
	Vector3 position() { return worldTransform.col3; }
	
	this()
	{
		_worldTransform = Matrix34(1);
		_invWorldTransform = Matrix34(1);
	}
}

class SceneObject : SceneObjectBase
{
	Mesh mesh;
	Material material;
	
	AABB getAABB()
	{
		if (mesh)
		{
			return mesh.getAABB(worldTransform);
		}
		else
		{
			return AABB(Vector3(0));
		}
	}
	
	void traceRay(RayContact * rc)
	{
		if (mesh is null) return;
		
		RayContact rcLocal;
		rcLocal.ray = rc.ray;
		rcLocal.ray.origin = mul(invWorldTransform, rc.ray.origin);
		rcLocal.ray.direction = rotate(invWorldTransform, rc.ray.direction);

		mesh.traceRay(&rcLocal);
		
		if (rcLocal.distance < rc.distance)
		{
			assert(rc.ray.insideRange(rcLocal.distance));
			
			rc.distance = rcLocal.distance; 
			rc.normal = normalize(rotate(worldTransform, rcLocal.normal));
			rc.ray.maxDistance = rcLocal.distance;
			rc.object = this;
		}
	}
}
