/**
 * delight.SceneSceneLoader
 * 
 * License: GPL
 * Author: Pyo Taesu
 */
module delight.SceneLoader;
private import delight.Scene;
private import delight.Math;
private import delight.Mesh;
private import delight.Material;
private import delight.Camera;
private import delight.Light;
private import delight.BIH;
private import tango.core.Exception;
private import tango.text.xml.Document;
private import tango.text.Util;
private import tango.io.File;
private import tango.io.Stdout;
private import Float = tango.text.convert.Float;
private import Integer = tango.text.convert.Integer;

class SceneLoader
{
	Scene scene;
	
	this(Scene scene)
	{
		this.scene = scene;
	}
	
	alias Document!(char) Doc;
	alias Doc.Node Node;

	static float stringToValue(T:float)(char[] s)
	{
		return Float.toFloat(trim(s));
	}
	
	static int stringToValue(T:int)(char[] s)
	{
		return Integer.toInt(trim(s));
	}
	
	static bool stringToValue(T:bool)(char[] s)
	{
		if (s == "true") return true;
		return false;
	}
	
	static T[] stringToValues(T)(char[] s)
	{
		auto l = split(s, ",");
		auto vals = new T[l.length];
		foreach( i, e ; l )
		{
			vals[i] = stringToValue!(T)(e);
		}
		return vals;
	}
	struct ForeachChildNodes
	{
		Node node;
		char[] name;
		int opApply(int delegate(ref Node) dg)
		{
			int r = 0;
			for(auto n = node.firstChild ; n ; n = n.nextSibling)
			{
				if (n.name == name)
				{
					r = dg(n);
					if (r) break;
				}
			}
			return r;
		}
	}
	
	ForeachChildNodes childNodes(Node node, char[] name)
	{
		ForeachChildNodes r = { node, name };
		return r;
	}
	
	Node findChildNode(Node node, char[] name)
	{
		foreach(n ; childNodes(node, name))
		{
			return n;
		}
		return null;
	}
	
	char[] getAttribute(Node node, char[] name, char[] defVal = "")
	{
		auto n = node.getAttribute(name);
		if (n) return n.value;
		return defVal;
	}
	
	char[] getText(Node node, char[] name)
	{
		auto n = findChildNode(node, name);
		if (n)
		{
			auto t = n.data((Node n){return true;});
			if (t) return t.value;
		}
		return "";
	}

	T[] getValues(T)(Node node, char[] name)
	{
		return stringToValues!(T)(getText(node, name));
	}

	Vector3 getVector3(Node node, char[] name, Vector3 defVal = Vector3(0))
	{
		auto v = getValues!(float)(node, name);
		if (v.length >= 3) return Vector3(v);
		return defVal;
	}
	
	Matrix34 getMatrix34(Node node, char[] name)
	{
		auto v = getValues!(float)(node, name);
		return Matrix34(v);
	}

	T getValue(T)(Node node, char[] name, T defVal)
	{
		auto s = getText(node, name);
		if (s.length > 0) return stringToValue!(T)(s);
		return defVal;
	}

	void parseCamera(Node node)
	{
		auto pos = getVector3(node, "pos");
		auto target = getVector3(node, "target");
		auto up = getVector3(node, "up");
		auto fov = degreeToRadian(getValue(node, "fov", 45.0f));

		scene.camera = new Camera;
		scene.camera.initCamera(pos, target, up, fov);
	}
	
	void parseRenderingOption(Node node)
	{
		int[] v = stringToValues!(int)(getText(node, "screensize"));
		if (v.length >= 2)
		{
			scene.renderingOption.screenWidth = v[0];
			scene.renderingOption.screenHeight = v[1];
		}
		
		scene.renderingOption.traceDepth = getValue(node, "tracedepth", 3);
	}
	
	void parseLight(Node node)
	{
		auto pos = getVector3(node, "pos");
		auto color = getVector3(node, "color");

		auto light = new Light;
		light.name = getAttribute(node, "name");
		light.initLight(pos, color);
		
		scene.addLight(light);
	}
	
	void parseMaterial(Node node, ref Material mat)
	{
		mat.color = getVector3(node, "color", mat.color);
		mat.specular = getValue(node, "specular", mat.specular);
		mat.specularPower = getValue(node, "specularPower", mat.specularPower);
		mat.fresnel = getValue(node, "fresnel", mat.fresnel);
		mat.refractionIndex = getValue(node, "refractionIndex", mat.refractionIndex);
		mat.reflective = getValue(node, "reflective", mat.reflective);
		mat.refractive = getValue(node, "refractive", mat.refractive);
	}
	
	void parseMaterial(Node node)
	{
		Material mat;
		parseMaterial(node, mat);

		auto name = getAttribute(node, "name");
		parseMaterial(node, mat);
		
		scene.addMaterial(name, mat);
	}
	
	void parseGeometry(Node node)
	{
		auto mesh = new Mesh;
		
		mesh.name = getAttribute(node, "name");
		auto meshType = getAttribute(node, "type");

		switch(meshType)
		{
		case "":
		case "mesh":
			mesh.loadFromObjFile( getText(node, "file") );
			break;
		case "plane":
			mesh.buildPlane();
			break;
		default:
			throw new NoSuchElementException("mesh type not found:" ~ meshType);
			break;
		}
		
		if (getValue(node, "flipnormal", false))
		{
			mesh.flipTriangles();
		}
		
		scene.addMesh(mesh);
	}
	
	void parseObject(Node node)
	{
		auto obj = new SceneObject;
		obj.name = getAttribute(node, "name");
		
		auto n = findChildNode(node, "geometry");
		
		obj.mesh = scene.getMesh(getAttribute(n, "name"));
		obj.worldTransform = getMatrix34(node, "transform");
		
		auto matNode = findChildNode(node, "material");
		auto matName = getAttribute(matNode, "name");
		obj.material = scene.getMaterial(matName);
		parseMaterial(matNode, obj.material);
		
		scene.addObject(obj);
	}

	void loadFromXml(char[] fileName)
	{
		Stdout("loading: ")(fileName)("...").newline.flush;
		
		scope file = new File(fileName);
		auto xml = cast(char[])file.read();

		auto doc = new Doc;
		doc.parse(xml);
		
		auto s = findChildNode(doc.root, "scene");
		
		parseCamera(findChildNode(s, "camera"));
		parseRenderingOption(findChildNode(s, "rendering"));
		
		foreach( o ; childNodes(s, "light"))
		{
			parseLight(o);
		}

		foreach( o ; childNodes(s, "material"))
		{
			parseMaterial(o);
		}

		foreach( o ; childNodes(s, "geometry"))
		{
			parseGeometry(o);
		}

		foreach( o ; childNodes(s, "object"))
		{
			parseObject(o);
		}
		
		scene.buildBVH();
	}
}
