/**
 * delight_gui
 * 
 * License: GPL
 * Author: Pyo Taesu
 */
module delight_gui;

import delight.Renderer;
import delight.Image;
import delight.Scene;
import delight.SceneLoader;
import delight.Util;
import delight.Math;
import Math = tango.math.Math;
import tango.core.Thread;
import tango.io.Stdout;
import dwt.DWT;
import dwt.widgets.Display;
import dwt.widgets.Shell;
import dwt.widgets.Canvas;
import dwt.widgets.Listener;
import dwt.widgets.Event;
import dwt.layout.FillLayout;
import dwt.graphics.GC;
import dwt.graphics.Image;
import dwt.graphics.Rectangle;
import dwt.graphics.Color;
import dwt.dwthelper.Runnable;

pragma(lib, "advapi32");
pragma(lib, "comctl32");
pragma(lib, "comdlg32");
pragma(lib, "gdi32");
pragma(lib, "kernel32");
pragma(lib, "shell32");
pragma(lib, "ole32");
pragma(lib, "oleaut32");
pragma(lib, "olepro32");
pragma(lib, "oleacc");
pragma(lib, "user32");
pragma(lib, "usp10");
pragma(lib, "msimg32");
//pragma(lib, "opengl32");
//pragma(lib, "shlwapi");
//pragma(lib, "gdiplus"); // load dynamic
//pragma(lib, "uxtheme"); // load dynamic
pragma(lib, "zlib");

// -------------------------------------------------------------------------

alias delight.Image.Image RenderImage;
alias dwt.graphics.Image.Image ScreenImage;

class DelightWindow
{
	int width = 800;
	int height = 600;
	ScreenImage screenImage;
	
	Display display;
	Canvas canvas;
	
	Thread[] threads = [];
	RenderBucket[] renderBuckets;
	RenderBucket[] renderedBuckets;
	
	Renderer renderer;
	RenderImage image;
	
	this(Display disp, Canvas cv)
	{
		display = disp;
		canvas = cv;
	}

	public Rectangle drawImage(GC gc)
	{
		if (screenImage)
		{
			gc.drawImage (screenImage, 0, 0);
			return screenImage.getBounds();
		}
		
		return null;
	}
	
	public void update()
	{
		RenderBucket rb;
		
		while(true)
		{
			synchronized(this)
			{
				auto n = renderedBuckets.length;
				if (n)
				{
					rb = renderedBuckets[n-1];
					renderedBuckets.length = n-1;
				}
				else
				{
					break;
				}
			}
			
			Stdout("Processed", rb.left, rb.top).newline;
			
			Color c = new Color(display);
			
			GC gc = new GC(screenImage);
			
			foreach(y ; range(rb.top, rb.bottom))
			{
				foreach(x ; range(rb.left, rb.right))
				{
					Vector3 p = image.getPixel(x, y);
					
					c.init_(
						cast(int)(p.x * 256),
						cast(int)(p.y * 256),
						cast(int)(p.z * 256)
						);
					
					gc.setForeground(c);
					gc.drawPoint(x, y);
				}
			}
			
			gc.dispose();
			
			canvas.redraw(rb.left, rb.top, rb.right - rb.left, rb.bottom - rb.top, false);
		}
	}
	
	public void startRendering()
	{
		auto scene = new Scene;
		auto sceneLoader = new SceneLoader(scene);
		sceneLoader.loadFromXml("scene.xml");

		Stdout("Image:")(scene.renderingOption.screenWidth)(" x ")(scene.renderingOption.screenHeight).newline.flush;

		image = new RenderImage;
		renderer = new Renderer(scene, image);
		
		screenImage = new ScreenImage(display, image.width, image.height);
		
		const u = 64;
		RenderBucket rb;

		for(int y = 0 ; y < image.height ; y += u)
		{
			for(int x = 0 ; x < image.width ; x += u)
			{
				rb.left = x;
				rb.top = y;
				rb.right = Math.min(x + u, image.width);
				rb.bottom = Math.min(y + u, image.height);
				
				auto n = renderBuckets.length;
				renderBuckets.length = n + 1;
				renderBuckets[n] = rb;
			}
		}
		
		startThreads();
	}
	
	public synchronized void startThreads()
	{
		auto thr = new Thread(&renderThread);
		thr.start();
		
		threads ~= [thr];
	}
	
	public void stopThreads()
	{
		
	}
	
	public void renderThread()
	{
		RenderBucket rb;

		while(true)
		{
			synchronized(this)
			{
				auto n = renderBuckets.length;
				if (n)
				{
					rb = renderBuckets[n-1];
					renderBuckets.length = n-1;
				}
				else
				{
					break;
				}
			}

			render(&rb);
			
			synchronized(this)
			{
				auto n = renderedBuckets.length;
				renderedBuckets.length = n + 1;
				renderedBuckets[n] = rb;
			}
		}
	}
	
	public void render(RenderBucket * rb)
	{
		renderer.render(rb);
	}
}

//-------------------------------------------------------------------------

int main(char[][] args)
{
	auto display = new Display();
	auto shell = new Shell();
	
	shell.setLayout(new FillLayout);
	auto canvas = new Canvas(shell, DWT.NO_BACKGROUND);

	auto wnd = new DelightWindow(display, canvas);
	
	canvas.addListener(DWT.Paint,
		new class Listener
		{
			public void handleEvent (Event e)
			{
				auto gc = e.gc;
				int w, h;

				Rectangle rect = wnd.drawImage(gc);
				
				if (rect)
				{
					w = rect.width;
					h = rect.height;
				}
				
				Rectangle client = canvas.getClientArea();
				
				int marginWidth = client.width - w;
				
				if (marginWidth > 0)
				{
					gc.fillRectangle (w, 0, marginWidth, client.height);
				}
				
				int marginHeight = client.height - h;
				
				if (marginHeight > 0)
				{
					gc.fillRectangle (0, h, client.width, marginHeight);
				}
			}
		}
	);
	
	shell.pack();
	canvas.pack();

	shell.setSize(800, 600);
	shell.open();
	
	wnd.startRendering();
	
	const timerPeriod = 500;
	
	display.timerExec(timerPeriod,
		new class Runnable
		{
			public void run()
			{
				wnd.update();
				display.timerExec(timerPeriod, this);								
			}
		}
	);
	
	while(!shell.isDisposed())
	{
		if (!display.readAndDispatch())
		{
			display.sleep();
		}
	}
	
	wnd.stopThreads();
	
	display.dispose();
	
	return 0;
}
